// Stepnode.cpp
//
// Copyright (C) 2009-2013 Mike McCauley
// $Id: Stepnode.cpp,v 1.20 2015/08/25 02:22:45 mikem Exp mikem $

#include "Stepnode.h"
#include "NodeManager.h"

Stepnode::Stepnode(NodeManager* manager, uint8_t id) {
    _manager = manager;
    _id = id;
    _currentPos = 0;
    _targetPos = 0;
    _speed = 0.0;
    _maxSpeed = 1.0;
    _acceleration = 0.0;
    _sqrt_twoa = 1.0;
    _stepInterval = 0;
    _minPulseWidth = 1;
    _lastStepTime = 0;
    state_enabled = false;
    state_endstop = false;

    // NEW
    _n = 0;
    _c0 = 0.0;
    _cn = 0.0;
    _cmin = 1.0;
    _direction = DIRECTION_CCW;

    // Some reasonable default
    setAcceleration(1);
}

void Stepnode::moveTo(long absolute) {
    if (_targetPos != absolute)
    {
	_targetPos = absolute;
	computeNewSpeed();
	// compute new n?
    }
}

void Stepnode::move(long relative) {
    moveTo(_currentPos + relative);
}

long Stepnode::distanceToGo() {
    return _targetPos - _currentPos;
}

long Stepnode::targetPosition() {
    return _targetPos;
}

long Stepnode::currentPosition() {
    return _currentPos;
}

// Useful during initialisations or after initial positioning
// Sets speed to 0
void Stepnode::setCurrentPosition(long position) {
    _targetPos = _currentPos = position;
    _n = 0;
    _stepInterval = 0;
}

void Stepnode::computeNewSpeed() {
    long distanceTo = distanceToGo(); // +ve is clockwise from curent location

    long stepsToStop = (long)((_speed * _speed) / (2.0 * _acceleration)); // Equation 16

    if (distanceTo == 0 && stepsToStop <= 1) {
    	// We are at the target and its time to stop
    	_stepInterval = 0;
    	_speed = 0.0;
    	_n = 0;
    	return;
    }

    if (distanceTo > 0) {
    	// We are anticlockwise from the target
    	// Need to go clockwise from here, maybe decelerate now
    	if (_n > 0) {
    	    // Currently accelerating, need to decel now? Or maybe going the wrong way?
    	    if ((stepsToStop >= distanceTo) || _direction == DIRECTION_CCW)
    		_n = -stepsToStop; // Start deceleration
    	} else if (_n < 0) {
    	    // Currently decelerating, need to accel again?
    	    if ((stepsToStop < distanceTo) && _direction == DIRECTION_CW)
    		_n = -_n; // Start accceleration
    	}
    } else if (distanceTo < 0) {
    	// We are clockwise from the target
    	// Need to go anticlockwise from here, maybe decelerate
    	if (_n > 0) {
    	    // Currently accelerating, need to decel now? Or maybe going the wrong way?
    	    if ((stepsToStop >= -distanceTo) || _direction == DIRECTION_CW)
    		_n = -stepsToStop; // Start deceleration
    	} else if (_n < 0) {
    	    // Currently decelerating, need to accel again?
    	    if ((stepsToStop < -distanceTo) && _direction == DIRECTION_CCW)
    		_n = -_n; // Start accceleration
    	}
    }

    // Need to accelerate or decelerate
    if (_n == 0) {
    	// First step from stopped
    	_cn = _c0;
    	_direction = (distanceTo > 0) ? DIRECTION_CW : DIRECTION_CCW;
    } else {
    	// Subsequent step. Works for accel (n is +_ve) and decel (n is -ve).
    	_cn = _cn - ((2.0 * _cn) / ((4.0 * _n) + 1)); // Equation 13
    	_cn = max(_cn, _cmin);
    }
    _n++;
    _stepInterval = _cn;
    _speed = 1000000.0 / _cn;
    if (_direction == DIRECTION_CCW)
	_speed = -_speed;
}

// Implements steps according to the current step interval
// You must call this at least once per step
// returns true if a step occurred
bool Stepnode::runSpeed() {
    // Dont do anything unless we actually have a step interval
    if (!_stepInterval)
	   return false;

    unsigned long time = micros();
    unsigned long nextStepTime = _lastStepTime + _stepInterval;
    // Gymnastics to detect wrapping of either the nextStepTime and/or the current time
    if (   ((nextStepTime >= _lastStepTime) && ((time >= nextStepTime) || (time < _lastStepTime)))
	|| ((nextStepTime < _lastStepTime) && ((time >= nextStepTime) && (time < _lastStepTime))))
    {
	if (_direction == DIRECTION_CW)
	{
	    // Clockwise
	    _currentPos += 1;
	}
	else
	{
	    // Anticlockwise
	    _currentPos -= 1;
	}
	step(_currentPos);

	_lastStepTime = time;
	return true;
    }
    else
    {
	return false;
    }
}

// Run the motor to implement speed and acceleration in order to proceed to the target position
// You must call this at least once per step, preferably in your main loop
// If the motor is in the desired position, the cost is very small
// returns true if the motor is still running to the target position.
bool Stepnode::run() {
    if (runSpeed())
	   computeNewSpeed();
    return _speed != 0.0 || distanceToGo() != 0;
}

void Stepnode::setMaxSpeed(float speed) {
    if (_maxSpeed != speed) {
    	_maxSpeed = speed;
    	_cmin = 1000000.0 / speed;
    	// Recompute _n from current speed and adjust speed if accelerating or cruising
    	if (_n > 0) {
    	    _n = (long)((_speed * _speed) / (2.0 * _acceleration)); // Equation 16
    	    computeNewSpeed();
    	}
    }
}

float Stepnode::maxSpeed() {
    return _maxSpeed;
}

void Stepnode::setAcceleration(float acceleration) {
    if (acceleration == 0.0)
    	return;
    if (_acceleration != acceleration) {
    	// Recompute _n per Equation 17
    	_n = _n * (_acceleration / acceleration);
    	// New c0 per Equation 7, with correction per Equation 15
    	_c0 = 0.676 * sqrt(2.0 / acceleration) * 1000000.0; // Equation 15
    	_acceleration = acceleration;
    	computeNewSpeed();
    }
}

void Stepnode::setSpeed(float speed) {
    if (speed == _speed)
        return;
    speed = constrain(speed, -_maxSpeed, _maxSpeed);
    if (speed == 0.0) {
    	_stepInterval = 0;
    } else {
    	_stepInterval = fabs(1000000.0 / speed);
    	_direction = (speed > 0.0) ? DIRECTION_CW : DIRECTION_CCW;
    }
    _speed = speed;
}

float Stepnode::speed() {
    return _speed;
}

// Subclasses can override
void Stepnode::step(long step) {
    if (_direction) {
        _manager->send |= (0b01 << (2*_id));
    } else {
        _manager->send |= (0b10 << (2*_id));
    }
}

void Stepnode::setMinPulseWidth(unsigned int minWidth) {
    _minPulseWidth = minWidth;
}

void Stepnode::stop() {
    if (_speed != 0.0) {
    	long stepsToStop = (long)((_speed * _speed) / (2.0 * _acceleration)) + 1; // Equation 16 (+integer rounding)
    	if (_speed > 0)
    	    move(stepsToStop);
    	else
    	    move(-stepsToStop);
    }
}
